////////////////////////////////////////////////////////////////////////////////
// Filename: Automaton.cc                                                    //
// Purpose : header file for call 'Automaton'                                 //
// provides several functions to build a NFA                                  //
////////////////////////////////////////////////////////////////////////////////
// Author: Markus Keidl                                                       //
// Theme.: 2                                                                  //
// Group.: 5 (Gerald Anleitner, Stefan Seltzsam, Markus Keidl)                //
//                                                                            //
// Date of creation: 28.05.1996        Date of last change: 17.07.1996        //
////////////////////////////////////////////////////////////////////////////////
 
#include "Automaton.hh"
#include "expand.cc"    // external file because of other author

// default constructor
Automaton::Automaton() : start_state(NULL) {} 

// destructor
Automaton::~Automaton()
{
  for(Pix i = state_list.first() ; i != 0 ; state_list.next(i))
      delete state_list(i);
  delete start_state; 
}

// add a new State to the Automaton
void Automaton::add_state(State* new_state) 
{
  if (new_state != NULL) state_list.append(new_state);
}

// set the start-State of the A
void Automaton::set_start_state(State* new_st_state)
{
  if (new_st_state != NULL) start_state = new_st_state;
}

// add a Transition to the State source
void Automaton::add_transition(State* source_state, Transition* new_trans)
{
  if (source_state != NULL || new_trans != NULL)
    source_state->add_transition(new_trans);
}
				    
// checks the beginning of text_to_scan to be a part of the language of the A
void Automaton::match(Input_Stream& text_to_scan, Protocol& hits)
{
  if (start_state != NULL)
    {
      text_to_scan.first();
      if (!text_to_scan)
	start_state->send(hits);
				// min. one Input_Atom available      
    }
}


// reads the file line by line
int Automaton::create_from_regexp(const String& path)
{
  int linenr = 0;
  int no_error = true;
  String line("");
  // create start state of automaton
  State* start = new Unlabelled_State;
  set_start_state(start);
  Input_Atom epsilon("");

  // open file
  char ch;
  ifstream read_from(path);
  if (read_from == 0)	       
    {
      cerr << path << ": no such file found" << endl;
      return false;
    }
  while(!read_from.eof())
    {
      // read a full line, cut off the last characte (='\n')
      line = "";
      while (read_from.get(ch) && ch != '\n')
	line += ch;
      linenr++;
      
      int counter = 0;				  // index for position in String line
      int len = line.length();			  // length of String line
      String token("");
      String regexp("");

      // ignore WHITESPACES before TOKEN
      while (len > counter && (line[counter] == ' ' || line[counter] == '\t'))
	counter++;
      if (len <= counter) continue;		  // only if line is empty

      // read TOKEN
      for ( ; len > counter && line[counter] != ' ' && line[counter] != '\t' ; counter++)
	{
	  if ((line[counter] >= 'A' && line[counter] <= 'Z') || line[counter] == '_')
	    token += line[counter];
	  else 
	    {
	      no_error = false;
	      cerr << path << ":" << linenr << ": " << "illegal character in token identifier" << endl;
	    }
	}

      // ignore WHITESPACES
      while (len > counter && (line[counter] == ' ' || line[counter] == '\t'))
	counter++;
	
      // read regular expression
      regexp = line.after(counter-1);
	
      // check for errors
      if  (regexp == "" && token != "")
	{
	  cerr << path << ":" << linenr << ": " << "regular expression missing :-<" << endl;
	  no_error = false;
	}
      else
	{
	  if (expand_regexp(regexp) == false)
	    {
	      // error is reported
	      cerr << path << ":" << linenr << ": " << regexp << endl;
	      no_error = false;
	    }
	  else
	    {
	      // Test
	      if(!is_valid(regexp))
		{
		  cerr << path << ":" << linenr << ": " << regexp << endl;
		  no_error = false;
		  continue;
		}

	      // build small automaton for each regular expression
	      State* start_state = new Unlabelled_State;
	      State* end_state = new Unlabelled_End_State(token);
	      
	      Transition* to_start_state = new Labelled_Transition(start_state,epsilon);
	      add_state(start_state);
	      add_state(end_state);
	      add_transition(start,to_start_state);

	      if (!interpret_regexp(start_state,end_state,regexp))
		{
		  cerr << path << ":" << linenr << ": " << regexp << endl;
		  return false;
		}
	    }
	}
    }
  if (!read_from.eof())
    {
      cerr << path << ": error occured while reading file" << endl;
      return false;
    }
  return no_error;
}

int Automaton::interpret_regexp(State* start, State* ende, String& regexp)
{
  Input_Atom epsilon("");

  if (regexp.length() == 1 || (regexp.length() == 2 && regexp[0] == '\\'))
  //  atomar letter        OR  quoted atomar letter 
    {
      if (regexp[0] == '\\')
	if (regexp[1] == 'e')         // quoted e = epsilon
	  regexp = "";
	else
	  regexp = regexp[1];
      Input_Atom label(regexp);
      Transition* trans = new Labelled_Transition(ende,label);
      add_transition(start,trans);

      return true;
    }

  // analyse regexp, search for leading operator
  State* current_state = start;
  while (regexp.length() != 0)
    {
      String op = get_next_op(regexp);
      if (op == "")
	return false;
	
      if (regexp.length() > 0)
	{
	  switch (regexp[0])
	    {
	    case '*':
	    case '+':
	    case '?':
	      {
	        // create states for '*', '?' and '+'
		State* new_start = new Unlabelled_State;
		State* new_ende = new Unlabelled_State;
		Transition* to_new_start_1 = new Labelled_Transition(new_start,epsilon);
		Transition* to_ende_2 = new Labelled_Transition(ende,epsilon);
		add_state(new_start);
		add_state(new_ende);

		if (regexp[0] != '?')
		  {
		    Transition* to_new_start_2 = new Labelled_Transition(new_start,epsilon);
		    add_transition(new_ende,to_new_start_2);    
		  }

		if (regexp[0] != '+')
		  {
		    Transition* to_ende_1 = new Labelled_Transition(ende,epsilon);
		    add_transition(current_state,to_ende_1);
		  }

		add_transition(current_state,to_new_start_1);
		add_transition(new_ende,to_ende_2);

		if (!interpret_regexp(new_start,new_ende,op))
		  {
		    regexp = op;
		    return false;
		  }
		current_state = ende;
		regexp = regexp.after(0);
		break;
	      }	  
	    case '|':
	      {
	        if (!interpret_regexp(current_state,ende,op))
		  {
		    regexp = op;
		    return false;
		  }
		current_state = start;
		regexp = regexp.after(0);
		break;
	      }
	    default:
	      {
	        // create states for '.'
		State* new_start = new Unlabelled_State;
		add_state(new_start);

      		if (!interpret_regexp(current_state,new_start,op))
		  {
		    regexp = op;
		    return false;
		  }
		current_state = new_start;
	      }
	    }
	}
      else
	if (!interpret_regexp(current_state,ende,op))
	  {
	    regexp = op;
	    return false;
	  }
    }
  return true;
}

String Automaton::get_next_op(String& regexp)
{
  String op("");
  int i = 1;
  switch (regexp[0])
    {
    case '(':
      {
	int counter=1;
	// get op between ( and )
	for( ; counter != 0 ; i++)
	  {
	    if (regexp[i] == '\\') 
	      {	      
		i += 1;
		continue;
	      }
	    if (regexp[i] == '(') counter += 1;
	    if (regexp[i] == ')') counter -= 1;
	  }
	op = regexp.at(1,i-2);
	break;
      }
    case '\\':
      {
	op = String(regexp[0]) + String(regexp[i++]);
	break;
      }
    default:
      op = regexp[0];
    }

  if (int(regexp.length()) > i)   // conversion because type of 
                                  // regexp.length() is unsigned int
    {
      regexp = regexp.after(i-1);
      if ((regexp[0] == '*' || regexp[0] == '+' || regexp[0] == '?') && (regexp.length() != 1))
	{
	  op = "(" + op + ")" + regexp[0];
	  regexp = regexp.after(0);
	}
    }
  else
    regexp = "";

  return op;
}

bool Automaton::is_valid(String& expression)
{
  Regex r("a**");

  String regexp(expression);
  regexp.gsub(Regex("\\\\.",1),"-");		// substitute all \x with -
  if (regexp[0] == '|' || regexp[0] == '*' || regexp[0] == '+' || regexp[0] == '?')
    // makno ),|,*,+,? at beginning of regexp allowed
    {
      expression="no |, *, +, ? at beginning of regular expression allowed";
      return false;
    }
  if (regexp [regexp.length()-1] == '|') // no  | at end of reg expression allowed
    {
      expression="missing operand at end of regular expression";
      return false;
    }
  if (regexp.contains(Regex("([)|\\*\\+\\?]",1))) // no ),|,*,+,? after ( allowed
    {
      expression="missing operand after (";
      return false;
    }
  if (regexp.contains(Regex("|[)\\*\\+\\?]",1))) // no ),*,+,? after | allowed
    {
      expression="missing operand after |";
      return false;
    }
  if (regexp.contains(Regex("[\\*\\+\\?][\\*\\+\\?]",1)))
    {
      expression="missing operand after */+/?";
      return false;
    }

  // count the number of parenthesis'
  int parent_nr = 0;
  int i = 0;

  while (int(regexp.length()) > i)
    {
      if (regexp[i] == '(') parent_nr +=1;
      if (regexp[i++] == ')') parent_nr -=1;
      if (parent_nr < 0)
	{
	  expression = "parenthesis mismatch";
	  return false;
	}
    }
  if (parent_nr != 0)
    {
      expression = "parenthesis mismatch";
      return false;
    }

  return true;
} 

// saves the nfa in the Sgraph-format in the file path
bool Automaton::sgraph(const String& path)
{
  int state_nrs_size = 1;
  String opts("");
  String label("");

  // open file
  ofstream write_to(path);
  if (!write_to)
      return false;

  // write header of Sgraph-file
  write_to << "GRAPH \"" << path << "\" = DIRECTED" << endl;
  write_to << "{$" << endl;
  write_to << endl;
  write_to << "$}" << endl;

  // save states in an array to identify each state with a number
  State** state_nrs = new(State*[state_list.length()+1]);
  state_nrs[0] = start_state;
  for (Pix i = state_list.first() ; i != 0 ; state_list.next(i))
    state_nrs[state_nrs_size++] = state_list(i);
  
  // for each state write all transitions in file path
  for (int  i = 0 ; i < state_nrs_size ; i++)
    {
      // write state i to file path
      opts = "";
      if (state_nrs[i]->get_token() != "")
	opts = "EN FF_C \"red\" ";      //marks an Endnode
      if (i == 0)		// start state
	opts = "SN FF_C \"blue\" ";
      write_to << i << " {$ " << opts <<"$} \"" << state_nrs[i]->get_token() << "\"" << endl;
      
      // for all transitions of this state
      for (state_nrs[i]->first() ; state_nrs[i]->operator()() != 0 ; state_nrs[i]->next())
	{
	  // write number of targetstate in state_nrs
	  write_to << " " << number_of_state(&state_nrs[i]->operator()()->get_targetstate(),state_nrs,state_nrs_size);
	  // write label
	  label = state_nrs[i]->operator()()->get_label().get_Input_Atom();
	  if (label == "\"" || label == "\\" || label == "{")
	    label = "\\" + label;
	  write_to << " {$ $} \"" << label << "\"" << endl;
	}
      // write character for end of state (;)
      write_to << " ;" << endl;
    }
  write_to << "END" << endl;
  return true;
}

// returns the number of which_state in array
int Automaton::number_of_state(State* which_state, State** array, int size)
{
  for (int i = 0 ; i < size ; i++)
      if (array[i] == which_state)
	return i;
  return -1;
}






